#!/usr/bin/perl -w
#
# shinGETsu - P2P anonymous BBS
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: gateway.cgi,v 1.90 2004/05/11 07:16:54 fuktommy Exp $
#

use lib qw(..);
use Shingetsu::Config;
use Shingetsu::NodeList;
use Shingetsu::Cache;
use Shingetsu::CacheStat;
use Shingetsu::Gateway;
use Shingetsu::Message;
use Shingetsu::Signature;
use Shingetsu::Util;

umask $Shingetsu::Config::umask;

$datadir	= $Shingetsu::Config::datadir;
$menufile	= $Shingetsu::Config::menufile;
$root		= $Shingetsu::Config::root;
@types		= qw(list thread note);
$message	= $Shingetsu::Gateway::message;

#
# sub rutines
#
sub xlocaltime(;$) {
	my @buf = Shingetsu::Util::localtime(@_);
	return sprintf "%d/%02d/%02d %02d:%02d",
		       1900+$buf[5], $buf[4]+1, $buf[3], $buf[2], $buf[1];
}

#
# main routine
#

if ($ENV{REMOTE_ADDR} !~ $Shingetsu::Config::friendAddr) {
	print403();
	exit;
} elsif (defined $ENV{PATH_INFO}) {
	$path = $ENV{PATH_INFO};
	$path =~ s|^/||;
} else {
	$path = "";
}

if ($ENV{REQUEST_METHOD} eq "POST") {
	read(STDIN, $input, $ENV{CONTENT_LENGTH});
} else {
	$input = $ENV{QUERY_STRING};
}

if (($path eq "")||($path eq "title")) {
	printTitle();

} elsif ($path eq "list") {
	printList($menufile);

} elsif ($path eq "motd") {
	printMotd();

} elsif ($path =~ m|^list/([0-9A-Za-z_%.]+)$|) {
	printList($1);

} elsif ($path =~ m|^thread/([0-9A-Za-z_%.]+)$|) {
	printThread($1);

} elsif ($path =~ m|^note/([0-9A-Za-z_%.]+)$|) {
	$num = $1;
	%arg = args($input);
	if ((! defined $arg{mode}) || ($arg{mode} eq "")) {
		printNote($num);
	} elsif ($arg{mode} eq "edit") {
		editDialog($num, "note");
	} elsif ($arg{mode} eq "history") {
		printHistory($num, "note");
	} else {
		printNote($num);
	}

} elsif ($path =~ m|^attach/([0-9A-Za-z_%.]+)/([0-9A-Za-z_]+)/([0-9]+)\.([0-9A-Za-z_]*)$|) {
	printAttach($1, $2, $3, $4);

} elsif ($path eq "check") {
	checkSign();

} elsif ($path eq "new") {
	printHeader("$$message{new}");
	newElementForm();
	print "</body></html>\n";

} elsif ($path eq "index") {
	checkAdmin();
	printIndex();

} elsif ($path eq "changes") {
	checkAdmin();
	printChanges();

} elsif ($path eq "update") {
	checkAdmin();
	printUpdate();

} elsif ($path =~ /^search/) {
	my $query = $';
	checkAdmin();
	if ((defined $query) && ($query =~ m|^/(.+)|)) {
		$query = str_decode($1);
		printSearchResult($query);
	} elsif ((defined $input) && ($input ne "")) {
		%arg = args($input);
		printSearchResult($arg{query});
	} else {
		printHeader($$message{search});
		searchForm();
		print "</body></html>";
	}

} elsif ($path eq "edittrust") {
	checkAdmin();
	%arg = args($input);
	if (! defined $arg{trip}) {
		print404();
		exit;
	} elsif ($arg{button} eq "ADD") {
		Shingetsu::Signature::addlist($arg{trip});
	} elsif ($arg{button} eq "DELETE") {
		Shingetsu::Signature::dellist($arg{trip});
	}
	print302("$ENV{SCRIPT_NAME}/$arg{type}/" .
		str_encode(file_decode("$arg{type}_$arg{file}")));


} elsif ($path =~ m|^post/([0-9A-Za-z_%.]+)$|) {
	my $num = $1;
	my $id = post($num, $input);
	if ($num =~ /^note_/) {
		print302("$ENV{SCRIPT_NAME}/note/" . str_encode(file_decode($num)));
	} else {
		print302("$ENV{SCRIPT_NAME}/thread/" . str_encode(file_decode($num)) . "#r$id");
	}


} elsif ($path =~ m|^add/([0-9A-Za-z_%.]+)$|) {
	my $num = $1;
	my $id = margeRecord($num, $input);
	print302("$ENV{SCRIPT_NAME}/note/" . str_encode(file_decode($num)));

} elsif ($path eq "newelement") {
	%arg = args($input);
	xdie $$message{null_name} if ($arg{name} eq "");
	xdie $$message{bad_name}  if ($arg{name} =~ m|[/#]|);
	xdie $$message{bad_name}  if ($arg{name} =~ m|\]\]|);
	xdie $$message{null_type} if ($arg{type} eq "");
	print302("$ENV{SCRIPT_NAME}/$arg{type}/" . str_encode($arg{name}));

} elsif ($path =~ m|^newelement/([0-9A-Za-z_]+)$|) {
	my $num = $1;
	newElement($num, $input);
	print302("$ENV{SCRIPT_NAME}/list/" . str_encode(file_decode($num)));

} elsif ($path eq "delete") {
	checkAdmin();
	deleteDialog();

} elsif ($path eq "xdelete") {
	checkAdmin();
	%arg = args($input);
	if (! defined $arg{file}) {
		print404();
		exit;
	} elsif ((defined $arg{stamp})&&(defined $arg{id})) {
		deleteRecord($arg{file}, $arg{stamp}, $arg{id});
	} else {
		deleteFile($arg{file});
	}

} elsif (($path =~ m|^[A-Z]|) && (-x "tradgw.cgi")) {
	print302("${root}tradgw.cgi/$path?$ENV{QUERY_STRING}");

} else {
	print404();

}

#------------------------------------------------------------------#

#
# check admin address
#
sub checkAdmin {
	if ($ENV{REMOTE_ADDR} !~ $Shingetsu::Config::adminAddr) {
		print403();
		exit;
	}
}

#
# print CGI&HTML header
#
sub printHeader {
	my($title) = @_;
	my $ver = $Shingetsu::Config::version;
	print "Content-Type: text/html; charset=UTF-8\r\n",
	      "Content-Language: $$message{lang}\r\n",
	      "X-Shingetsu: shinGETsu/$ver\r\n",
	      "\r\n";
	print '<?xml version="1.0" encoding="UTF-8"?>',
	      '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"',
	      ' "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">',
	      "<html xmlns='http://www.w3.org/1999/xhtml' xml:lang='$$message{lang}'>",
	      '<head>',
	      "<meta http-equiv='Content-Language' content='$$message{lang}' />",
	      '<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />',
#	      "<meta http-equiv='Content-Style-Type' content='text/css' />",
	      "<title>$title</title>",
	      "<link rev='made' href='mailto:fuktommy\@users.sourceforge.net' />",
	      "<link rel='contents' href='$ENV{SCRIPT_NAME}/list' />",
	      "<link rel='stylesheet' type='text/css' href='$Shingetsu::Config::css' />",
#	      "<link rel='shortcut icon' href='/favicon.ico' />",
	      "</head>";
	print "<body><p>",
	      "<a href='$ENV{SCRIPT_NAME}'>$$message{top}</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/list'>$$message{menu}</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/new'>$$message{new}</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/index'>$$message{index}</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/changes'>$$message{changes}</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/update'>$$message{update}</a> | ",
	      "<a href='$ENV{SCRIPT_NAME}/search'>$$message{search}</a>",
	      "</p>",
	      "<h1><a href='$ENV{SCRIPT_NAME}/search/", str_encode($title),
	      "' class='title' title='$$message{search}'>$title</a></h1>"
}

#
# print CGI header (403 forbidden)
#
sub print403 {
	my ($file) = @_;
	printHeader($$message{403});
	print "</body></html>\n";
}

#
# print CGI header (404 not found)
#
sub print404 {
	my ($file) = @_;
	printHeader($$message{404});
	print "<p>$$message{try_later}</p>";
	if (defined $file) {
	      print "<form method='get' action='$ENV{SCRIPT_NAME}/delete?file=$file'>",
		    "<p><input type='submit' value='$$message{del_file}' tabindex='1' accesskey='d' /></p></form>";
	}
	print "</body></html>\n";
}

#
# print CGI header (302 moved temporarily)
#
sub print302 {
	my($next) = @_;
	my $ver = $Shingetsu::Config::version;
	my($title) = @_;
	print "Content-Type: text/html; charset=UTF-8\r\n",
	      "Content-Language: en\r\n",
	      "X-Shingetsu: shinGETsu/$ver\r\n",
	      "\r\n",
	      '<?xml version="1.0" encoding="UTF-8"?>',
	      '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"',
	      ' "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">',
	      '<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">',
	      '<head>',
	      '<meta http-equiv="Content-Language" content="en" />',
	      '<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />',
	      '<meta http-equiv="Content-Script-Type" content="text/javascript" />',
	      "<title>moved</title>",
	      "<link rev='made' href='mailto:fuktommy\@users.sourceforge.net' />",
	      "<link rel='contents' href='$ENV{SCRIPT_NAME}/list' />",
	      "</head>",
	      "<body><p>Click and jump to <a href='$next'>$next</a></p>\n",
	      "<script type='text/javascript'>window.location.href = '$next';</script>",
	      "</body></html>\n";
}

#
# print title page
#
sub printTitle {
	printHeader($$message{logo});
	print "<p>$$message{license}</p>\n",
	      "<ul><li><em><a href='$ENV{SCRIPT_NAME}/list'>$$message{menu}</a></em></li>\n";
	print "<li><a href='${root}tradgw.cgi'>",
	      "$$message{tradgw}</a></li>\n" if (-x "tradgw.cgi");
	print "<li><a href='http://shingetsu.sourceforge.net/'>",
	      "$$message{site}</a></li>\n",
	      "<li><a href='$ENV{SCRIPT_NAME}/motd'>$$message{agreement}</a></li></ul>",
	      "</body></html>\n";
}

#
# print MOTD file
#
sub printMotd {
	print "Content-type: text/plain; charset=euc-jp\r\n",
	      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n\r\n";
	open IN, $Shingetsu::Config::motd;
	while (<IN>) {
		print;
	}
	close IN;
}

#
# print list
#
sub printList {
	my($list) = @_;
	local($_, @_);
	$list = str_decode($list);
	my $file_list = file_encode("list", $list);
	touch($file_list);
	open IN, "$datadir/$file_list.dat" or die "$datadir/$file_list.dat: failed to open. $!";
	printHeader($list);
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete'>",
	      "<p><input type='hidden' name='file' value='$file_list' /></p><ul>";
	my @buf = ();
	my %stamp = ();
	my %stat = Shingetsu::CacheStat::list();
	my $count = 0;
	while (<IN>) {
		$count++;
		my %rec = rec($_);
		next if (keys %rec <= 2);
		$rec{body} = "" unless (defined $rec{body});
		my $stamp = $rec{stamp};
		my $buf;
		if ($rec{type}) {
			my $file_child = file_encode($rec{type}, $rec{link});
			my $file = "$datadir/$file_child.dat";
			my $records = 0;
			my $size    = 0;
			if (-s $file) {
				$stamp   = $stat{$file_child}{stamp};
				$records = $stat{$file_child}{records};
				$size	 = int($stat{$file_child}{size}/1024/1024*10)/10;
			}
			$records = "?" if ($records <= 0);
			my $str = xlocaltime($stamp);
			my $child = str_encode($rec{link});
			$buf = "<li><input type='radio' name='record' value='$rec{stamp}/$rec{id}' " .
			       "tabindex='1' accesskey='s' />" .
			       "$str: <a href='$ENV{SCRIPT_NAME}/$rec{type}/$child'>$rec{link}</a>" . 
			       "($$message{$rec{type}}/$records/$size$$message{mb})$rec{body}</li>\n";
		} elsif ($rec{remove_stamp}) {
			$stamp = $rec{remove_stamp};
			my $str = xlocaltime($stamp);
			my $check = "";
			$rec{name} = "" unless (defined $rec{name});
			if ($rec{pubkey}) {
				$check = "<a href='$ENV{SCRIPT_NAME}/check?" .
					"file=" . file_encode("list", $list) . "&amp;" .
					"stamp=$rec{stamp}&amp;id=$rec{id}'>" .
					Shingetsu::Signature::pubkey2trip($rec{pubkey}) . "</a>";
			}
			$buf = "<li><input type='radio' name='record' value='$rec{stamp}/$rec{id}' " .
			       "tabindex='1' accesskey='s' />" .
			       "$str: $rec{name} $check $$message{remove}:$rec{remove} $rec{body}</li>\n";
		}
		push @buf, $buf;
		$stamp{$buf} = $stamp;
	}
	close IN;
	@buf = sort {$stamp{$b} <=> $stamp{$a}} @buf;
	print @buf;
	print "</ul><p><input type='submit' value='$$message{del_record}' tabindex='2' accesskey='d' />",
	      "<input type='hidden' name='mode' value='list' /></p></form>";
	my $filesize = int((-s "$datadir/$file_list.dat")/1024/1024*10)/10;
	my $checked = ($count)? "checked='checked'": "";
	if ($filesize <= $Shingetsu::Config::filelimit) {
		newElementForm(file=>$file_list, checked=>$checked, dopost=>1);
	}
	print "<form action='$ENV{SCRIPT_NAME}/delete'>",
	      "<p><input type='submit' value='$$message{del_file}' tabindex='6' accesskey='d' />",
	      "<input type='hidden' name='file' value='$file_list' />",
	      " ${filesize}$$message{mb}</p></form></body></html>\n";
}

#
# format to HTML
#
sub html_format {
	local($_) = @_;
	s|<br>|<br />|g;
	s|(&gt;&gt;)([0-9A-Za-z]+)|<a href='#r$2'>$&</a>|g;
	s|https?://[^\x00-\x20"'()<>\[\]\x7F-\xFF]{2,}|<a href='$&'>$&</a>|g;
	s|\[\[<a.*?>(.*?)\]\]</a>|[[$1]]|g;
	s|\[\[([^<>]+?)\]\]|sprintf("<a href='%s'>[[%s]]</a>",str_encode($1),$1)|eg;
	if ($Shingetsu::Config::redirect) {
		s|<a href='http://shingetsu\.p2p/|<a href='/|g;
	}
	return $_;
}

#
# print thread
#
sub printThread {
	my($thread) = @_;
	local $_;
	$thread = str_decode($thread);
	my $file_thread = file_encode("thread", $thread);
	touch($file_thread);
	open IN, "$datadir/$file_thread.dat" or die "$datadir/$file_thread.dat: failed to open. $!";
	printHeader($thread);
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete'>",
	      "<p><input type='hidden' name='file' value='$file_thread'  /></p><dl>\n";
	my $count = 0;
	while (<IN>) {
		$count++;
		%rec = rec($_);
		next if (keys %rec <= 2);
		my $xstamp = xlocaltime $rec{stamp};
		$rec{name} = $$message{anonymous} if ((! defined $rec{name}) || ($rec{name} eq ""));
		my $mail = "";
		$mail = "[$rec{mail}]" if (defined $rec{mail});
		$rec{body} = "" if (! defined $rec{body});
		$rec{body} = html_format($rec{body});
		if ($Shingetsu::Config::redirect) {
			$rec{body} =~ s|<a href='http://shingetsu\.p2p/|<a href='/|g;
		}
		my $attach = "";
		my $check  = "";
		my $remove = "";
		if ($rec{attach}) {
			my $size = int(length($rec{attach}) / 4 * 3 / 1024);
			$attach = " <a href='$ENV{SCRIPT_NAME}/attach/" .
				  str_encode($thread) . "/$rec{id}/$rec{stamp}.$rec{suffix}'>" .
				  "$rec{stamp}.$rec{suffix}</a> (${size}$$message{kb})";
		}
		if ($rec{pubkey}) {
			$check = "<a href='$ENV{SCRIPT_NAME}/check?" .
				 "file=" . file_encode("thread", $thread) . "&amp;" .
				 "stamp=$rec{stamp}&amp;id=$rec{id}'>" .
				 Shingetsu::Signature::pubkey2trip($rec{pubkey}) . "</a>";
		}
		if (($rec{remove_id}) && ($rec{remove_stamp})) {
			my $xid = substr($rec{remove_id},0,8);
			$remove = "[[$$message{remove}: <a href='#r$xid'>$xid</a>]]<br />";
		}
		my $xid = substr $rec{id}, 0, 8;
		print "<dt id='r$xid'><input type='radio' name='record' value='$rec{stamp}/$rec{id}' ",
		      "tabindex='1' accesskey='s' />",
		      "$xid <span class='name'>$rec{name}</span>",
		      " $mail $check $xstamp$attach</dt><dd>$rec{body}<br />$remove<br /></dd>\n";
	}
	close IN;
	print "</dl><p><input type='submit' value='$$message{del_record}' tabindex='2' accesskey='d' />",
      	      "<input type='hidden' name='mode' value='thread' /></p></form>";
	my $filesize = int((-s "$datadir/$file_thread.dat")/1024/1024*10)/10;
	my $checked = ($count)? "checked='checked'": "";
	my $readonly = ($ENV{REMOTE_ADDR} =~ $Shingetsu::Config::adminAddr)? "": "readonly='readonly'";
	if ($filesize <= $Shingetsu::Config::filelimit) {
		print "<form method='post' action='$ENV{SCRIPT_NAME}/post/$file_thread' enctype='multipart/form-data'><p>",
		"<input type='submit' value='$$message{post}' name='submit' tabindex='1' accesskey='w' />",
		" <input type='checkbox' value='dopost' name='dopost'",
		" $checked tabindex='2' accesskey='s' />",
		"$$message{send}",
		" <input type='checkbox' value='error' name='error'",
		" checked='checked' tabindex='2' accesskey='e' />",
		"$$message{error}<br />",
		" $$message{name}:<input name='name' size='15' value='' tabindex='3' accesskey='n' />",
		" $$message{mail}:<input name='mail' size='15' value='' tabindex='4' accesskey='m' />",
		" $$message{signature}:<input type='password' name='passwd' size='15' $readonly value='' tabindex='5' accesskey='p' /><br />",
		" $$message{attach}:<input type='file' name='attach' size='19' value='' tabindex='6' accesskey='a' />",
		" $$message{suffix}:<select name='suffix' size='1' tabindex='7'>",
		"<option>AUTO</option>";
		foreach (sort (keys %Shingetsu::Config::mimeType)) {
			print "<option>$_</option>";
		}
		print "</select> ($$message{limit}: ${Shingetsu::Config::filelimit}$$message{mb})<br />",
		"<textarea rows='5' cols='70' name='body' tabindex='8' accesskey='c'></textarea>",
		"</p></form>";
	}
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete'>",
	      "<p><input type='submit' value='$$message{del_file}' tabindex='9' accesskey='d' />",
	      "<input type='hidden' name='file' value='$file_thread' />",
	      " ${filesize}$$message{mb}</p></form></body></html>\n";
}

#
# print note
#
sub printNote {
	my($note) = @_;
	local $_;
	$note = str_decode($note);
	my $file_note = file_encode("note", $note);
	touch($file_note);
	printHeader($note);
	my($buf, $conflict) = readNote($file_note);
	my %rec = rec($buf);
	$rec{body} = "" unless (defined $rec{body});
	my $body = html_format($rec{body});
	print "<p>$body</p>";
	print "<p><em>$$message{conflict}</em></p>" if ($conflict);
	my $filesize = int((-s "$datadir/$file_note.dat")/1024/1024*10)/10;
	if ($filesize <= $Shingetsu::Config::filelimit) {
		print "<form method='post' action='$ENV{SCRIPT_NAME}/add/$file_note'><p>",
		"<input type='submit' value='$$message{add}' name='submit' tabindex='1' accesskey='w' />",
		" <input type='checkbox' value='dopost' name='dopost'",
		" checked='checked' tabindex='2' accesskey='s' />",
		"$$message{send}",
		" <input type='checkbox' value='error' name='error'",
		" checked='checked' tabindex='2' accesskey='e' />",
		"$$message{error}<br />",
		"<textarea rows='5' cols='70' name='message' tabindex='8' accesskey='c'></textarea>",
		"<br /><a href='$ENV{SCRIPT_NAME}/note/", str_encode($note), "?mode=edit'>$$message{edit}</a>",
		" | <a href='$ENV{SCRIPT_NAME}/note/", str_encode($note), "?mode=history'>$$message{history}</a>",
		"</p></form>";
	} else {
		print "<p><a href='$ENV{SCRIPT_NAME}/note/", str_encode($note), "?mode=history'>$$message{history}</a></p>";
	}
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete'>",
	      "<p><input type='submit' value='$$message{del_file}' tabindex='9' accesskey='d' />",
	      "<input type='hidden' name='file' value='$file_note' />",
	      " ${filesize}$$message{mb}</p></form></body></html>\n";
}

#
# print attachment
#
sub printAttach {
	my($thread, $id, $stamp, $suffix) = @_;
	$thread = file_encode("thread", str_decode($thread));
	my $ver = $Shingetsu::Config::version;
	$suffix = lc $suffix;
	my $type = $Shingetsu::Config::mimeType{$suffix};
	$type = "text/plain" unless (defined $type);
	touch($thread);
	open IN, "$datadir/$thread.dat" or die "$datadir/$thread.dat: failed to open. $!";
	my $value;
	while (<IN>) {
		my %rec = rec($_);
		if ($rec{attach} && ($rec{stamp} eq $stamp) && ($rec{id} eq $id)) {
			$value = $rec{attach};
			last;
		}
	}
	close IN;
	if (defined $value) {
		print "Content-Type: $type\r\n",
		      "X-Shingetsu: shinGETsu/$ver\r\n",
		      "\r\n",
		       Shingetsu::Extern::base64decode($value);
	} else {
		print404($thread);
	}
}

#
# check signature
#
sub checkSign {
	my %arg = args($input);
	unless ($arg{file} && $arg{stamp} && $arg{id}) {
		print404();
		exit;
	}
	my $file = $arg{file};
	touch($file);
	open IN, "$datadir/$file.dat" or die "$datadir/$file.dat: failed to open. $!";
	my $value;
	while (<IN>) {
		my %rec = rec($_);
		next unless (($rec{stamp} eq $arg{stamp}) && ($rec{id} eq $arg{id}));
		if (! ((defined $rec{pubkey}) && ($rec{pubkey} ne "")
			&& (defined $rec{sign}) && ($rec{sign} ne "")
			&& (defined $rec{target}) && ($rec{target} ne ""))) {
			print "Content-Type: text/plain; charset=UTF-8\r\n",
			      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n",
			      "\r\n",
			      "$$message{not_signed}\n$_\n";
		} elsif (Shingetsu::Signature::check(%rec)) {
			s/</&lt;/g; s/>/&gt;/g;
			my $trip = Shingetsu::Signature::pubkey2trip($rec{pubkey});
			my($type, $filename) = split /_/, $file, 2;
			printHeader($$message{check_sign});
			print "<p>$$message{protected}($rec{target}).</p>",
			"<form method='get' action='$ENV{SCRIPT_NAME}/edittrust'><p>",
			"<input type='submit' name='button' value='$$message{add}' tabindex='1' accesskey='a' />",
			"/<input type='submit' name ='button' value='$$message{delete}' tabindex='2' accesskey='d' />",
			"<input type='hidden' name='trip' value='$trip' />",
			"<input type='hidden' name='type' value='$type' />",
			"<input type='hidden' name='file' value='$filename' />",
			"$$message{to_from_trust_list}: $trip</p></form>",
			"<p>$_</p>",
			"</body></html>";
			close IN;
			exit;
		} else {
			print "Content-Type: text/plain; charset=UTF-8\r\n",
			      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n",
			      "\r\n",
			      "$$message{wrong_sign}.\n$_\n";
			close IN;
			exit;
		}
	}
	close IN;
	print404($file);
}

#
# print index page
#
sub printIndex {
	printHeader($$message{index});
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete'>\n";
	local $_;
	my $type;
	my %stat = Shingetsu::CacheStat::list();
	foreach $type (@types) {
		print "<h2>$$message{$type}</h2><ul>\n";
		foreach (sort glob("$datadir/${type}_*.dat")) {
			$_ =~ s|^$datadir/||;	$_ =~ s|.dat$||;
			my $x = file_decode($_);
			my $y = str_encode($x);
			my $records = $stat{$_}{records};
			my $size    = int($stat{$_}{size}/1024/1024*10)/10;
			print "<li><input type='radio' name='file' value='$_' tabindex='1' accesskey='s' />",
			      "<a href='$ENV{SCRIPT_NAME}/$type/$y'>$x</a>($records/$size$$message{mb})</li>\n";
		}
		print "</ul>"
	}
	print "<p><input type='submit' value='$$message{del_file}' tabindex='2' accesskey='d' /></p></form>";
	newElementForm();
	print "</body></html>\n";
}

#
# print recent changes
#
sub printChanges {
	printHeader($$message{changes});
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete'>\n";
	local $_;
	my $type;
	my %stat = Shingetsu::CacheStat::list();
	foreach $type (@types) {
		print "<h2>$$message{$type}</h2><ul>\n";
		my @buf = ();
		my %stamp = ();
		foreach (glob("$datadir/${type}_*.dat")) {
			$_ =~ s|^$datadir/||;	$_ =~ s|.dat$||;
			my $x = file_decode($_);
			my $y = str_encode($x);
			my $date = xlocaltime($stat{$_}{stamp});
			my $records = $stat{$_}{records};
			my $size    = int($stat{$_}{size}/1024/1024*10)/10;
			my $buf = "<li><input type='radio' name='file' value='$_' tabindex='1' accesskey='s' />$date <a href='$ENV{SCRIPT_NAME}/$type/$y'>$x</a>($records/$size$$message{mb})</li>\n";
			$stamp{$buf} = $stat{$_}{stamp};
			push @buf, "$buf";
		}
		@buf = sort {$stamp{$b} <=> $stamp{$a}} @buf;
		print @buf;
		print "</ul>"
	}
	print "<p><input type='submit' value='$$message{del_file}' tabindex='2' accesskey='d' /></p></form>";
	newElementForm();
	print "</body></html>\n";
}

#
# print recent update
#
sub printUpdate {
	local $_;
	my %stamp = ();
	if (open IN, $Shingetsu::Config::updatelist) {
		while (<IN>) {
			chomp;
			@_ = split /<>/;
			if ($_[2] =~ /^(list|thread|note)_/) {
				$stamp{$_[2]} = $_[0];
			}
		}
		close IN;
	} else {
		print404($$message{update});
		exit;
	}
	printHeader($$message{update});
	my $type;
	my %stat = Shingetsu::CacheStat::list();
	foreach $type (@types) {
		print "<h2>$$message{$type}</h2><ul>\n";
		my @buf = ();
		foreach (sort {$stamp{$b}<=>$stamp{$a}} grep(/^${type}_/, keys(%stamp))) {
			my $x = file_decode($_);
			my $y = str_encode($x);
			my $date = xlocaltime $stamp{$_};
			my $records = (defined $stat{$_})? $stat{$_}{records}: "?";
			my $size    = (defined $stat{$_})? int($stat{$_}{size}/1024/1024*10)/10: "?";
			print "<li>$date <a href='$ENV{SCRIPT_NAME}/$type/$y'>$x</a>($records/$size$$message{mb})</li>\n";
		}
		print "</ul>"
	}
	newElementForm();
	print "</body></html>\n";
}

#
# edit form
#
sub editForm($$) {
	my($file, $rec) = @_;
	my $body = (defined $$rec{body})? $$rec{body}: "";
	$body =~ s|<br>|\n|g;
	print "<form method='post' action='$ENV{SCRIPT_NAME}/post/$file' enctype='multipart/form-data'><p>",
	      "<input type='submit' value='$$message{post}' name='submit' tabindex='1' accesskey='w' />",
	      " <input type='checkbox' value='dopost' name='dopost'",
	      " checked='checked' tabindex='2' accesskey='s' />",
	      "$$message{send}",
	      " <input type='checkbox' value='error' name='error'",
	      " checked='checked' tabindex='2' accesskey='e' />",
	      "$$message{error}<br />",
	      "<textarea rows='20' cols='75' name='body' tabindex='8' accesskey='c'>$body</textarea>",
	      "<input type='hidden' value='$$rec{stamp}' name='base_stamp' />",
	      "<input type='hidden' value='$$rec{id}' name='base_id' />",
	      "</p></form>";
}

#
# print note history
#
sub printHistory {
	my($name, $type) = @_;
	local $_;
	$name = str_decode($name);
	my $file = file_encode($type, $name);
	touch($file);
	open IN, "$datadir/$file.dat" or die "$datadir/$file.dat: failed to open. $!";
	printHeader($name);
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete'>",
	      "<p><input type='hidden' name='file' value='$file'  /></p><dl>\n";
	my @buf = <IN>;
	my %ref = ();
	my $last = "";
	foreach (@buf) {
		my %rec = rec($_);
		next if (keys %rec <= 2);
		$ref{"$rec{stamp}_$rec{id}"} = 0;
		if (! $rec{base_stamp}) {
			$ref{"$rec{stamp}_$rec{id}"} = 1;
			next;
		} elsif (defined $ref{"$rec{base_stamp}_$rec{base_id}"}) {
			$ref{"$rec{base_stamp}_$rec{base_id}"}++;
		}
		$last = $_;
	}
	if ($last) {
		my %rec = rec($last);
		$ref{"$rec{stamp}_$rec{id}"} = 1;
	}
	my $last_record;
	foreach (@buf) {
		my %rec = rec($_);
		next if (keys %rec <= 2);
		my $xstamp = xlocaltime $rec{stamp};
		$rec{body} = "" if (! defined $rec{body});
		$rec{body} = html_format($rec{body});
		my $base = "";
		if ($rec{base_id}) {
			$base = substr $rec{base_id}, 0, 8;
			$base =~ s/"/&quot;/g;
			$base = "(<a href=\"#r$base\">&gt;&gt;$base</a>)" if ($base);
		}
		my $remove = "";
		if (($rec{remove_id}) && ($rec{remove_stamp})) {
			my $xid = substr($rec{remove_id},0,8);
			$remove = "[[$$message{remove}: <a href='#r$xid'>$xid</a>]]<br />";
		}
		my $xid = substr $rec{id}, 0, 8;
		my $conflict = ($ref{"$rec{stamp}_$rec{id}"})? "": " <em>$$message{conflict}</em>";
		print "<dt id='r$xid'><input type='radio' name='record' value='$rec{stamp}/$rec{id}' ",
		      "tabindex='1' accesskey='s' />",
		      "$xid $xstamp $base$conflict</dt><dd>$rec{body}<br />$remove<br /></dd>\n";
	}
	close IN;
	print "</dl><p><input type='submit' value='$$message{del_record}' tabindex='2' accesskey='d' />",
      	      "<input type='hidden' name='mode' value='$type' /></p></form>";
	my($buf, $conflict) = readNote($file);
	my %rec = rec($buf);
	editForm($file, \%rec);
	my $filesize = int((-s "$datadir/$file.dat")/1024/1024*10)/10;
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete'>",
	      "<p><input type='submit' value='$$message{del_file}' tabindex='9' accesskey='d' />",
	      "<input type='hidden' name='file' value='$file' />",
	      " ${filesize}$$message{mb}</p></form></body></html>\n";
}

#
# edit dialog
#
sub editDialog {
	my($name, $type) = @_;
	$name = str_decode($name);
	my $file = file_encode($type, $name);
	touch($file);
	my($buf, $conflict) = readNote($file);
	my %rec = rec($buf);
	$rec{body} = "" unless (defined $rec{body});
	my $xbody = html_format($rec{body});
	printHeader($name);
	print "<p>$xbody</p>\n";
	print "<p><em>$$message{conflict}</em></p>" if ($conflict);	
	editForm($file, \%rec);
	my $filesize = int((-s "$datadir/$file.dat")/1024/1024*10)/10;
	print "<form method='get' action='$ENV{SCRIPT_NAME}/delete'>",
	      "<p><input type='submit' value='$$message{del_file}' tabindex='9' accesskey='d' />",
	      "<input type='hidden' name='file' value='$file' />",
	      " ${filesize}$$message{mb}</p></form></body></html>\n";
}

#
# delete record form
#
sub deleteRecordForm($$$$$) {
	my($file, $buf, $stamp, $id, $arg) = @_;
	$buf =~ s/</&lt;/g;
	$buf =~ s/>/&gt;/g;
	print	"<p>$$message{del_record_q}</p>",
		"<form method='post' action='$ENV{SCRIPT_NAME}/xdelete'><p>",
		"<input type='hidden' value='$file' name='file' />",
		"<input type='hidden' value='$stamp' name='stamp' />",
		"<input type='hidden' value='$id' name='id' />",
		"<input type='submit' value='$$message{yes}' name='submit' tabindex='1' accesskey='w' />",
		" <input type='checkbox' value='dopost' name='dopost' tabindex='2' accesskey='s' />",
		"$$message{send}",
		" <input type='checkbox' value='error' name='error'",
		" checked='checked' tabindex='2' accesskey='e' />",
		"$$message{error}<br />",
		" $$message{name}:<input name='name' size='15' value='' tabindex='3' accesskey='n' />",
		" $$message{signature}:<input type='password' name='passwd' size='15' value='' tabindex='4' accesskey='p' />",
		" $$message{comment}:<input name='message' size='15' value='' tabindex='5' accesskey='m' />",
		"<input type='hidden' name='mode' value='$$arg{mode}' />",
		"</p></form>\n",
		"<p>$buf</p>";
}

#
# delete file form
#
sub deleteFileForm($@) {
	my($file, @buf) = @_;
	print "<p>$$message{del_file_q} ",
	      "<a href='$ENV{SCRIPT_NAME}/xdelete?file=$file'>",
	      "$$message{yes}</a></p>\n";
	if ($file =~ /_/) {
		print "<p>", file_decode($file), "($$message{$`})</p>\n";
	}
	foreach (@buf) {
		chomp;
		s/</&lt;/g;
		s/>/&gt;/g;
		print "<p>$_</p>\n";
	}
}

#
# delete record dialog
#
sub deleteRecordDialog(%) {
	my %arg = @_;
	printHeader($$message{del_record});
	my $file = $arg{file};
	print "\n";
	my($stamp, $id) = split /\//, $arg{record};
	my $buf;
	if (open IN, "$datadir/$file.dat") {
		while (<IN>) {
			if (/^$stamp<>$id<>/) {
				$buf = $_;
				last;
			}
		}
		close IN;
		if (! defined $buf) {
			print "<p>$$message{no_record}</p></body></html>\n";
		} else {
			deleteRecordForm($file, $buf, $stamp, $id, \%arg);
			print "</body></html>\n";
		}
	} else {
		print "<p>$file: $$message{no_file}</p></body></html>\n";
	}
}

#
# delete file dialog
#
sub deleteFileDialog(%) {
	my %arg = @_;
	my $file = $arg{file};
	printHeader("$$message{del_file}");
	print "\n";
	my @buf = ();
	my $count = 0;
	if (open IN, "$datadir/$file.dat") {
		while (<IN>) {
			push @buf, $_;
			last if ($count++ > 1);
		}
		close IN;
		deleteFileForm($file, @buf);
		print "</body></html>\n";
	} else {
		print "<p>$file: $$message{no_file}</p></body></html>\n";
	}
}

#
# delete dialog
#
sub deleteDialog {
	my %arg = args($input);
	local $_;
	if (! defined $arg{file}) {
		print404();
		exit;
	} elsif (defined $arg{record}) {
		deleteRecordDialog(%arg);
	} else {
		deleteFileDialog(%arg);
	}
}

#
# delete file
#
sub deleteFile {
	my($file) = @_;
	if (unlink "$datadir/$file.dat") {
		my %stat = Shingetsu::CacheStat::list();
		delete $stat{$file};
		Shingetsu::CacheStat::sync(%stat);
		print302("$ENV{SCRIPT_NAME}/list");
	} else {
		print "Content-Type: text/plain\r\n\r\n",
		      "failed. $!\n";
	}
}

#
# delete record
#
sub deleteRecord {
	my($file, $stamp, $id) = @_;
	my %arg = args($input);
	local $_;
	if (! Shingetsu::Cache::removeRecord($file, $stamp, $id)) {
		print "Content-Type: text/plain\r\n\r\n",
		      "failed. $!\n";
	} elsif ($arg{dopost}) {
		postDeleteMessage($file, $stamp, $id, %arg);
	}
	if ((defined $arg{mode}) && ($arg{mode} ne "")) {
		my $xfile = str_encode(file_decode($file));
		print302("$ENV{SCRIPT_NAME}/$arg{mode}/$xfile");
	} else {
		print "Content-Type: text/plain\r\n\r\n",
		      "succeed. $!\n";
	}
}

#
# new element form
# $opt{file}, $opt{checked}, $opt{dopost}
#
sub newElementForm {
	my %opt = @_;
	$opt{checked} = "" unless ($opt{checked});
	my $file = (defined $opt{file})? "/$opt{file}": "";

	print "<form method='get' action='$ENV{SCRIPT_NAME}/newelement$file'><p>",
	      "<input type='submit' value='$$message{new}' name='submit' tabindex='1' accesskey='w' />",
	      "<input type='radio' value='list' name='type' tabindex='2' accesskey='t' />",
	      "$$message{list}",
	      " <input type='radio' value='thread' name='type' tabindex='2' accesskey='t' />",
	      "$$message{thread}",
	      " <input type='radio' value='note' name='type' tabindex='2' accesskey='t' />",
	      "$$message{note}";
	print " <input type='checkbox' value='dopost' name='dopost'",
	      " $opt{checked} tabindex='2' accesskey='s' />",
	      "$$message{send}" if ($opt{dopost});
	print "<br />$$message{name}: <input name='name' size='19' value='' tabindex='3' accesskey='t' />",
	      "</p></form>";
}

#
# search form
#
sub searchForm {
	my $query = (@_)? $_[0]: "";
	my $xquery = $query;
	$xquery =~ s/"/&quot;/g;
	print "<form method='get' action='$ENV{SCRIPT_NAME}/search'><p>",
	      "<input name='query' size='19' value=\"$xquery\" tabindex='1' accesskey='q' />",
	      "<input type='submit' value='$$message{search}' name='submit' tabindex='1' accesskey='s' />",
	      "</p></form>";

	my $gquery = $query;
	$gquery =~ s/\|/ OR /g;
	$gquery =~ s/\.[+*]?/ /g;
	$gquery =~ s/"/&quot;/g;
	print "<form method='get' action='$Shingetsu::Config::google'><p>",
	      "<input name='q' size='19' value=\"$gquery\" tabindex='1' accesskey='g' />",
	      "<input type='hidden' name='ie' value='UTF-8' />",
	      "<input type='submit' name='google' value='$$message{google}' tabindex='1' accesskey='g' />",
	      "</p></form>";

}

#
# search result
#
sub printSearchResult {
	my($query) = @_;
	printHeader($$message{search});
	print "\n";
	local $_;
	my($type, $file);
	my %stat = Shingetsu::CacheStat::list();
	foreach $type (@types) {
		print "<h2>$$message{$type}</h2><ul>\n";
		my @file = search($type, $query);
		foreach $file (@file) {
			my $x = file_decode($file);
			my $y = str_encode($x);
			my $records = $stat{$file}{records};
			my $size    = int($stat{$file}{size}/1024/1024*10)/10;
			print "<li><a href='$ENV{SCRIPT_NAME}/$type/$y'>$x</a>($records/$size$$message{mb})</li>\n";
		}
		print "</ul>"
	}
	searchForm($query);
	print "</body></html>";
}
